package de.hft.stuttgart.citydoctor2.healer.gui;

import java.util.ArrayList;
import java.util.EnumMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;

import de.hft.stuttgart.citydoctor2.check.Check;
import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.CheckId;
import de.hft.stuttgart.citydoctor2.check.RequirementType;
import de.hft.stuttgart.citydoctor2.check.Checker;
import de.hft.stuttgart.citydoctor2.check.Requirement;
import de.hft.stuttgart.citydoctor2.datastructure.AbstractBuilding;
import de.hft.stuttgart.citydoctor2.datastructure.BoundarySurface;
import de.hft.stuttgart.citydoctor2.datastructure.Building;
import de.hft.stuttgart.citydoctor2.datastructure.Installation;
import de.hft.stuttgart.citydoctor2.datastructure.BuildingPart;
import de.hft.stuttgart.citydoctor2.datastructure.CityDoctorModel;
import de.hft.stuttgart.citydoctor2.datastructure.CityObject;
import de.hft.stuttgart.citydoctor2.datastructure.ConcretePolygon;
import de.hft.stuttgart.citydoctor2.datastructure.Geometry;
import de.hft.stuttgart.citydoctor2.datastructure.GeometryType;
import de.hft.stuttgart.citydoctor2.datastructure.GmlId;
import de.hft.stuttgart.citydoctor2.datastructure.LinkedPolygon;
import de.hft.stuttgart.citydoctor2.datastructure.Lod;
import de.hft.stuttgart.citydoctor2.datastructure.Opening;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;
import de.hft.stuttgart.citydoctor2.gui.AutoProWindow;
import de.hft.stuttgart.citydoctor2.gui.HighlightController;
import de.hft.stuttgart.citydoctor2.gui.ListErrorVisitor;
import de.hft.stuttgart.citydoctor2.gui.TriangulatedGeometry;
import de.hft.stuttgart.citydoctor2.healer.Healer;
import de.hft.stuttgart.citydoctor2.optimization.MeshGenerator;
import de.hft.stuttgart.citydoctor2.utils.Copy;
import javafx.collections.ObservableList;
import javafx.scene.control.TextInputDialog;
import javafx.scene.control.TreeItem;
import javafx.scene.shape.CullFace;
import javafx.scene.shape.DrawMode;

public class HealerController {

	private CityDoctorModel model;
	private HealerView healerView;

	private CityObject currentFeature;
	private Geometry currentGeometry;

	private CityObject nextFeature;
	private Geometry nextGeometry;

	private Checker checker;
	private CullFace culling = CullFace.BACK;
	private DrawMode drawMode = DrawMode.FILL;

	private TriangulatedGeometry nextTriangulatedGeometry;
	private TriangulatedGeometry currentTriangulatedGeometry;

	private ListErrorVisitor currentErrorVisitor;
	private ListErrorVisitor nextErrorVisitor;

	private HighlightController currentHighlights;
	private HighlightController nextHighlights;

	private SolidInjectorDialog injectorDialog;

	public HealerController(HealerView healerView) {
		this.healerView = healerView;
		this.currentHighlights = new HighlightController(healerView.getCurrentWorld());
		this.nextHighlights = new HighlightController(healerView.getNextWorld());
		currentErrorVisitor = new ListErrorVisitor(currentHighlights);
		nextErrorVisitor = new ListErrorVisitor(nextHighlights);
		healerView.translateZProperty().addListener((obs, oldV, newV) -> {
			currentHighlights.changeScaling(-newV.doubleValue());
			nextHighlights.changeScaling(-newV.doubleValue());
		});
	}
	
	public void showSimplyfication() {
		if (model == null) {
			return;
		}
		model.createFeatureStream().forEach(CityObject::prepareForChecking);
		AutoProWindow window = new AutoProWindow(model);
		window.setLocationRelativeTo(null);
		window.setVisible(true);
	}

	public void setCurrentModel(CityDoctorModel model) {
		if (this.model != model) {
			healerView.getCurrentMeshGroup().getChildren().clear();
			healerView.getNextMeshGroup().getChildren().clear();
			healerView.getCurrentErrorList().getItems().clear();
		}
		this.model = model;
	}

	public void selectFeatureClicked() {
		if (model == null) {
			return;
		}
		ChooseFeatureDialog dialog = healerView.getChooseFeatureDialog();
		dialog.getBuildingsTree().getRoot().getChildren().clear();
		model.getBuildings().forEach(b -> {
			if (hasBuildingGeometryErrors(b)) {
				addBuildingToTree(b, dialog);
			}
		});
		dialog.show();
	}

	private boolean hasBuildingGeometryErrors(Building b) {
		boolean hasError = hasAbstractBuildingGeometryErrors(b);
		if (hasError) {
			return true;
		}
		for (BuildingPart bp : b.getBuildingParts()) {
			if (hasAbstractBuildingGeometryErrors(bp)) {
				return true;
			}
		}
		return false;
	}

	private boolean hasAbstractBuildingGeometryErrors(AbstractBuilding b) {
		for (Geometry geom : b.getGeometries()) {
			if (geom.containsAnyError()) {
				return true;
			}
		}
		if (doBoundarySurfaceHaveGeometryErrors(b.getBoundarySurfaces())) {
			return true;
		}
		return doBuildingInstallationsContainGeometryErrors(b.getBuildingInstallations());
	}

	private void addBuildingToTree(Building b, ChooseFeatureDialog dialog) {
		TreeItem<SelectableFeature> buildingItem = new TreeItem<>(new SelectableFeature(b.getGmlId().getGmlString()));
		dialog.getBuildingsTree().getRoot().getChildren().add(buildingItem);
		buildingItem.setExpanded(true);
		buildGeometryTreeItems(b, buildingItem);
		buildBoundarySurfaceItems(b.getBoundarySurfaces(), buildingItem);
		buildBuildingInstallationItems(b.getBuildingInstallations(), buildingItem);
	}

	private void buildBuildingInstallationItems(List<Installation> buildingInstallations,
			TreeItem<SelectableFeature> buildingItem) {
		boolean hasBiError = doBuildingInstallationsContainGeometryErrors(buildingInstallations);
		if (hasBiError) {
			TreeItem<SelectableFeature> allInstItem = new TreeItem<>(new SelectableFeature("BuildingInstallations"));
			buildingItem.getChildren().add(allInstItem);
			for (Installation bi : buildingInstallations) {
				if (!hasBuildingInstallationGeometryError(bi)) {
					continue;
				}
				TreeItem<SelectableFeature> instItem = new TreeItem<>(
						new SelectableFeature(bi.getGmlId().getGmlString()));
				allInstItem.getChildren().add(instItem);
				buildBoundarySurfaceItems(bi.getBoundarySurfaces(), instItem);
			}
		}
	}

	private boolean doBuildingInstallationsContainGeometryErrors(List<Installation> buildingInstallations) {
		for (Installation bi : buildingInstallations) {
			boolean hasError = hasBuildingInstallationGeometryError(bi);
			if (hasError) {
				return true;
			}
		}
		return false;
	}

	private boolean hasBuildingInstallationGeometryError(Installation bi) {
		for (Geometry geom : bi.getGeometries()) {
			if (geom.containsAnyError()) {
				return true;
			}
		}
		return doBoundarySurfaceHaveGeometryErrors(bi.getBoundarySurfaces());
	}

	private void buildBoundarySurfaceItems(List<BoundarySurface> boundarySurfaces, TreeItem<SelectableFeature> root) {
		boolean hasBsError = doBoundarySurfaceHaveGeometryErrors(boundarySurfaces);
		if (hasBsError) {
			TreeItem<SelectableFeature> surfacesItem = new TreeItem<>(new SelectableFeature("BoundarySurfaces"));
			root.getChildren().add(surfacesItem);
			// dummy
			TreeItem<SelectableFeature> dummyItem = new TreeItem<>(null);
			surfacesItem.getChildren().add(dummyItem);

			// expand listener to actually load the boundary surfaces
			// this is for saving memory
			surfacesItem.expandedProperty().addListener((obs, oldV, newV) -> {
				if (Boolean.TRUE.equals(newV) && surfacesItem.getChildren().size() == 1) {
					surfacesItem.getChildren().clear();
					for (BoundarySurface bs : boundarySurfaces) {
						if (hasBoundarySurfaceGeometryError(bs)) {
							TreeItem<SelectableFeature> surfaceItem = new TreeItem<>(
									new SelectableFeature(bs.getGmlId().getGmlString() + " [" + bs.getType() + "]"));
							surfacesItem.getChildren().add(surfaceItem);
							buildGeometryTreeItems(bs, surfaceItem);
						}
					}
				}
			});
		}
	}

	private boolean hasBoundarySurfaceGeometryError(BoundarySurface bs) {
		for (Geometry geom : bs.getGeometries()) {
			if (geom.containsAnyError()) {
				return true;
			}
		}
		return false;
	}

	private void buildGeometryTreeItems(CityObject co, TreeItem<SelectableFeature> root) {
		for (Geometry geom : co.getGeometries()) {
			if (!geom.containsAnyError()) {
				continue;
			}
			TreeItem<SelectableFeature> geomItem = new TreeItem<>(
					new SelectableGeometry(createGeometryText(geom), geom, co));
			root.getChildren().add(geomItem);
		}
	}

	private boolean doBoundarySurfaceHaveGeometryErrors(List<BoundarySurface> boundarySurfaces) {
		for (BoundarySurface bs : boundarySurfaces) {
			boolean hasError = hasBoundarySurfaceGeometryError(bs);
			if (hasError) {
				return true;
			}
		}
		return false;
	}

	private String createGeometryText(Geometry geom) {
		StringBuilder sb = new StringBuilder();
		sb.append("Geometry [");
		sb.append(geom.getType());
		sb.append(", ");
		sb.append(geom.getLod());
		if (geom.getGmlId().isGenerated()) {
			sb.append("]");
		} else {
			sb.append("] ");
			sb.append(geom.getGmlId());
		}
		return sb.toString();
	}

	public void itemSelected(TreeItem<SelectableFeature> item, ChooseFeatureDialog dialog) {
		if (item == null) {
			return;
		}
		SelectableFeature value = item.getValue();
		dialog.setOkButtonEnabled(value.isGeometry());
	}

	public void setCurrentChecker(Checker checker) {
		this.checker = checker;
	}

	public void featureSelectionConfirmed(SelectableFeature selectedItem) {
		if (!(selectedItem instanceof SelectableGeometry)) {
			return;
		}
		SelectableGeometry selectedGeometry = (SelectableGeometry) selectedItem;
		currentGeometry = selectedGeometry.getGeometry();
		currentFeature = selectedGeometry.getFeature();
		currentTriangulatedGeometry = TriangulatedGeometry.of(currentGeometry);
		currentErrorVisitor.setGeometry(currentTriangulatedGeometry);
		currentTriangulatedGeometry.setCullFace(culling);
		currentTriangulatedGeometry.setDrawMode(drawMode);
		healerView.setSelectedFeatureText(currentFeature.getGmlId().getGmlString());
		healerView.setFeatureTypeText(currentFeature.getFeatureType().toString());
		healerView.setGeometryText(currentGeometry.getLod().toString() + ", " + currentGeometry.getType().toString());
		updateCurrentErrors();
		healerView.zoomOutForBoundingBox(currentGeometry.calculateBoundingBox());
		healerView.getCurrentMeshGroup().getChildren().clear();
		healerView.getNextMeshGroup().getChildren().clear();
		healerView.getCurrentMeshGroup().getChildren().addAll(currentTriangulatedGeometry.getMeshes());
		healerView.getNextStepBtn().setDisable(false);
		healerView.getHealCompleteBtn().setDisable(false);
		healerView.getAcceptBtn().setDisable(true);
		healerView.getCancelBtn().setDisable(true);
	}

	private void updateCurrentErrors() {
		List<CheckError> errors = new ArrayList<>();
		currentGeometry.collectContainedErrors(errors);
		healerView.getCurrentErrorList().getItems().clear();
		for (CheckError err : errors) {
			healerView.getCurrentErrorList().getItems().add(err);
		}
	}

	public void nextStepClicked() {
		if (currentGeometry == null || checker == null || currentFeature == null) {
			return;
		}
		healerView.getHealCompleteBtn().setDisable(true);
		currentFeature.prepareForChecking();
		if (nextGeometry == null) {
			GmlId id = currentGeometry.getGmlId();
			nextFeature = Copy.copy(currentFeature);
			nextGeometry = findNextGeometry(id, nextFeature);
		}
		nextFeature.prepareForChecking();
		Healer healer = new Healer(checker);
		healer.heal(nextGeometry, nextFeature, 1);
		nextFeature.prepareForChecking();
		checker.executeChecksForCheckable(nextFeature);
		updateNextErrors();
		updateNextGeometryView();
	}

	private void updateNextGeometryView() {
		nextTriangulatedGeometry = TriangulatedGeometry.of(nextGeometry);
		nextErrorVisitor.setGeometry(nextTriangulatedGeometry);
		nextTriangulatedGeometry.setCullFace(culling);
		nextTriangulatedGeometry.setDrawMode(drawMode);
		healerView.getNextMeshGroup().getChildren().clear();
		healerView.getNextMeshGroup().getChildren().addAll(nextTriangulatedGeometry.getMeshes());
		healerView.getAcceptBtn().setDisable(false);
		healerView.getCancelBtn().setDisable(false);
		healerView.getSolidInjectorBtn().setDisable(true);
	}

	public void healCompleteClicked() {
		if (currentGeometry == null) {
			return;
		}
		// generate a gml id if there is none, so the selected geometry can be found
		// later
		// needs to be done befoe copy, otherwise copy will copy null
		GmlId gmlId = currentGeometry.getGmlId();
		nextFeature = Copy.copy(currentFeature);
		nextGeometry = findNextGeometry(gmlId, nextFeature);
		Healer healer = new Healer(checker);
		healer.heal(nextGeometry, nextFeature);
		updateNextErrors();
		nextTriangulatedGeometry = TriangulatedGeometry.of(nextGeometry);
		nextErrorVisitor.setGeometry(nextTriangulatedGeometry);
		nextTriangulatedGeometry.setCullFace(culling);
		nextTriangulatedGeometry.setDrawMode(drawMode);
		healerView.getNextMeshGroup().getChildren().clear();
		healerView.getNextMeshGroup().getChildren().addAll(nextTriangulatedGeometry.getMeshes());
		healerView.getAcceptBtn().setDisable(false);
		healerView.getNextStepBtn().setDisable(true);
		healerView.getHealCompleteBtn().setDisable(true);
		healerView.getCancelBtn().setDisable(false);
	}

	private void updateNextErrors() {
		List<CheckError> errors = new ArrayList<>();
		nextGeometry.collectContainedErrors(errors);
		ObservableList<CheckError> items = healerView.getNextErrorList().getItems();
		items.clear();
		for (CheckError err : errors) {
			items.add(err);
		}
	}

	private Geometry findNextGeometry(GmlId gmlId, CityObject nextFeature) {
		Geometry[] foundGeom = new Geometry[1];
		Check geomFinder = new Check() {

			@Override
			public void check(Geometry geom) {
				if (geom.getGmlId().equals(gmlId)) {
					foundGeom[0] = geom;
				}
			}

			@Override
			public RequirementType getType() {
				return null;
			}

			@Override
			public Check createNewInstance() {
				return null;
			}

			@Override
			public CheckId getCheckId() {
				return null;
			}

			@Override
			public Set<Requirement> appliesToRequirements() {
				return null;
			}
		};
		nextFeature.accept(geomFinder);
		return foundGeom[0];
	}

	public void acceptHealingClicked() {
		if (model == null || checker == null || nextFeature == null || currentFeature == null) {
			return;
		}
		nextFeature.prepareForChecking();
		checker.executeChecksForCheckable(nextFeature);
		model.replaceFeature(currentFeature, nextFeature);
		currentFeature = nextFeature;
		currentGeometry = nextGeometry;
		updateCurrentErrors();
		healerView.getNextErrorList().getItems().clear();
		healerView.getCurrentMeshGroup().getChildren().clear();
		healerView.getCurrentMeshGroup().getChildren().addAll(nextTriangulatedGeometry.getMeshes());
		currentTriangulatedGeometry = nextTriangulatedGeometry;
		currentErrorVisitor.setGeometry(currentTriangulatedGeometry);
		nextTriangulatedGeometry = null;
//		nextHighlights = new HighlightController(healerView.getNextWorld());
//		nextErrorVisitor = new ListErrorVisitor(nextHighlights);
		healerView.getNextMeshGroup().getChildren().clear();
		healerView.getHealCompleteBtn().setDisable(false);
		healerView.getNextStepBtn().setDisable(false);
		healerView.getAcceptBtn().setDisable(true);
		healerView.getCancelBtn().setDisable(true);
	}

	public void setWireframe(boolean b) {
		if (b) {
			drawMode = DrawMode.LINE;
		} else {
			drawMode = DrawMode.FILL;
		}
		if (currentTriangulatedGeometry != null) {
			currentTriangulatedGeometry.setDrawMode(drawMode);
		}
		if (nextTriangulatedGeometry != null) {
			nextTriangulatedGeometry.setDrawMode(drawMode);
		}
	}

	public void setCulling(CullFace culling) {
		this.culling = culling;
		if (currentTriangulatedGeometry != null) {
			currentTriangulatedGeometry.setCullFace(culling);
		}
		if (nextTriangulatedGeometry != null) {
			nextTriangulatedGeometry.setCullFace(culling);
		}
	}

	public void cancelClicked() {
		if (model == null || nextFeature == null) {
			return;
		}
		healerView.getNextErrorList().getItems().clear();
		healerView.getNextMeshGroup().getChildren().clear();
		healerView.getAcceptBtn().setDisable(true);
		healerView.getCancelBtn().setDisable(true);
		healerView.getNextStepBtn().setDisable(false);
		healerView.getHealCompleteBtn().setDisable(false);
		nextFeature = null;
		nextGeometry = null;
		nextTriangulatedGeometry = null;
	}

	public void currentHighlight(CheckError err) {
		if (err == null) {
			currentHighlights.clearHighlights();
			return;
		}
		err.accept(currentErrorVisitor);
	}

	public void nextHighlight(CheckError err) {
		if (err == null) {
			nextHighlights.clearHighlights();
			return;
		}
		err.accept(nextErrorVisitor);
	}

	public void onHide() {
		currentHighlights.clearHighlights();
		nextHighlights.clearHighlights();
		healerView.getCurrentErrorList().getSelectionModel().clearSelection();
		healerView.getNextErrorList().getSelectionModel().clearSelection();
		healerView.getSelectFeatureBtn().setDisable(true);
	}

	public void onShow(CityDoctorModel model, Checker checker) {
		if (model != null) {
			// solid injector and meshing works without validation first
			this.model = model;
			healerView.getSolidInjectorBtn().setDisable(false);
			healerView.getMeshBtn().setDisable(false);
		} else {
			healerView.getSolidInjectorBtn().setDisable(true);
			healerView.getMeshBtn().setDisable(true);
		}
		
		// TODO: refactor this better
		if (model == null || checker == null) {
			resetGUI();
			healerView.getSelectFeatureBtn().setDisable(true);
			healerView.getFixBtn().setDisable(true);
			healerView.getHealerToolBar().getSaveBtn().setDisable(true);
			return;
		}
		if (model != this.model || checker != this.checker) {
			this.model = model;
			this.checker = checker;
			resetGUI();
			healerView.getSelectFeatureBtn().setDisable(false);
			healerView.getFixBtn().setDisable(false);
			healerView.getHealerToolBar().getSaveBtn().setDisable(false);
		}
	}

	private void resetGUI() {
		healerView.getAcceptBtn().setDisable(true);
		healerView.getCancelBtn().setDisable(true);
		healerView.getNextStepBtn().setDisable(true);
		healerView.getHealCompleteBtn().setDisable(true);
		healerView.getCurrentErrorList().getItems().clear();
		healerView.getNextErrorList().getItems().clear();
		healerView.getCurrentMeshGroup().getChildren().clear();
		healerView.getNextMeshGroup().getChildren().clear();
		healerView.setSelectedFeatureText("");
		healerView.setFeatureTypeText("");
		healerView.setGeometryText("");
	}

	public void injectSolid(boolean useBs, boolean useBi, boolean useLod2, boolean useLod3, boolean useLod4) {
		if (model == null || currentFeature == null) {
			return;
		}
		CityObject oldFeature = currentFeature;
		currentFeature.prepareForChecking();
		if (currentFeature instanceof BoundarySurface) {
			BoundarySurface surface = (BoundarySurface) currentFeature;
			currentFeature = surface.getParent();
		}
		if (!(currentFeature instanceof Building)) {
			return;
		}
		// collect polygons
		Map<Lod, List<Polygon>> availablePolygons = new EnumMap<>(Lod.class);
		Map<Lod, Set<Polygon>> existingPolygons = new EnumMap<>(Lod.class);
		Building b = (Building) currentFeature;
		collectPolygons(useBs, useBi, availablePolygons, existingPolygons, b);
		if (useLod2) {
			insertSolidGeometry(availablePolygons, b, existingPolygons, Lod.LOD2);
		}
		if (useLod3) {
			insertSolidGeometry(availablePolygons, b, existingPolygons, Lod.LOD3);
		}
		if (useLod4) {
			insertSolidGeometry(availablePolygons, b, existingPolygons, Lod.LOD4);
		}
		nextFeature = null;
		nextGeometry = null;
		Lod lod = currentGeometry.getLod();
		GeometryType type = GeometryType.SOLID;
		currentGeometry = Objects.requireNonNull(currentFeature.getGeometry(type, lod));
		checker.executeChecksForCheckable(currentFeature);
		healerView.getNextErrorList().getItems().clear();
		healerView.getNextMeshGroup().getChildren().clear();
		updateCurrentErrors();
		currentTriangulatedGeometry = TriangulatedGeometry.of(currentGeometry);
		currentErrorVisitor.setGeometry(currentTriangulatedGeometry);
		currentTriangulatedGeometry.setCullFace(culling);
		currentTriangulatedGeometry.setDrawMode(drawMode);
		healerView.zoomOutForBoundingBox(currentGeometry.calculateBoundingBox());

		healerView.getCurrentMeshGroup().getChildren().clear();
		healerView.getCurrentMeshGroup().getChildren().addAll(currentTriangulatedGeometry.getMeshes());

		healerView.getNextStepBtn().setDisable(false);
		healerView.getHealCompleteBtn().setDisable(false);
		healerView.getAcceptBtn().setDisable(true);
		healerView.getCancelBtn().setDisable(true);
		healerView.getSolidInjectorBtn().setDisable(true);
		currentFeature = oldFeature;
	}

	private void collectPolygons(boolean useBs, boolean useBi, Map<Lod, List<Polygon>> availablePolygons,
			Map<Lod, Set<Polygon>> existingPolygons, Building b) {
		if (useBs) {
			collectPolygons(availablePolygons, b.getBoundarySurfaces());
		}
		if (useBi) {
			for (Installation bi : b.getBuildingInstallations()) {
				for (Geometry geom : bi.getGeometries()) {
					List<Polygon> polygons = availablePolygons.computeIfAbsent(geom.getLod(), l -> new ArrayList<>());
					polygons.addAll(geom.getPolygons());
				}
				collectPolygons(availablePolygons, bi.getBoundarySurfaces());
			}
		}
		for (Geometry geom : b.getGeometries()) {
			Set<Polygon> polygons = existingPolygons.computeIfAbsent(geom.getLod(), l -> new HashSet<>());
			polygons.addAll(geom.getPolygons());
		}
	}

	private void insertSolidGeometry(Map<Lod, List<Polygon>> availablePolygons, Building b,
			Map<Lod, Set<Polygon>> existingPolygons, Lod lod) {
		List<Polygon> newPolygons = availablePolygons.get(lod);
		if (newPolygons != null) {
			// found polygons to be added
			Geometry newGeometry = new Geometry(GeometryType.SOLID, lod);
			Set<Polygon> polygons = existingPolygons.get(lod);
			if (polygons != null) {
				// add existing polygons
				for (Polygon p : polygons) {
					newGeometry.addPolygon(p);
				}
			}
			for (Polygon p : newPolygons) {
				if (p instanceof ConcretePolygon && !newGeometry.containsPolygon(p)) {
					LinkedPolygon newPoly = new LinkedPolygon((ConcretePolygon) p, newGeometry);
					newGeometry.addPolygon(newPoly);
				}
			}
			// remove existing geometry, if possible
			b.removeGeometry(lod);
			// add new geometry
			b.addGeometry(newGeometry);
			newGeometry.updateEdgesAndVertices();
			if (currentGeometry == null) {
				currentGeometry = newGeometry;
			}
		}

	}

	private void collectPolygons(Map<Lod, List<Polygon>> availablePolygons, List<BoundarySurface> surfaces) {
		for (BoundarySurface bs : surfaces) {
			for (Geometry geom : bs.getGeometries()) {
				List<Polygon> polygons = availablePolygons.computeIfAbsent(geom.getLod(), l -> new ArrayList<>());
				polygons.addAll(geom.getPolygons());
			}
			for (Opening o : bs.getOpenings()) {
				for (Geometry geom : o.getGeometries()) {
					List<Polygon> polygons = availablePolygons.computeIfAbsent(geom.getLod(), l -> new ArrayList<>());
					polygons.addAll(geom.getPolygons());
				}
			}
		}
	}

	public void solidInjectorClicked() {
		if (injectorDialog == null) {
			injectorDialog = new SolidInjectorDialog(this);
		}
		injectorDialog.show();
	}

	public void fixModel() {
		if (model == null || checker == null) {
			return;
		}
		Healer healer = new Healer(checker);
		healer.healModel(model);
	}

	public void meshBtnClicked() {
		if (model == null) {
			return;
		}
		TextInputDialog dialog = new TextInputDialog("1");
		dialog.setTitle("Mesh Parameter Input");
		dialog.setHeaderText("Input a number > 0.01");
		dialog.setContentText("Maximum Area of meshed triangles in m²:");

		Optional<String> result = dialog.showAndWait();
		if (result.isPresent()){
			String areaString = result.get();
			try {
				double area = Double.parseDouble(areaString);
				if (area < 0.01) {
					// triangle area is too small, ignore
					return;
				}
				for (Building b : model.getBuildings()) {
					MeshGenerator.replaceGeometryWithMeshedGeometry(b, area);
				}
			} catch (NumberFormatException e) {
				// no real number was inserted, do nothing
			}
		}
	}

}
